/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * This file is part of SableCC.                             *
 * See the file "LICENSE" for copyright information and the  *
 * terms and conditions for copying, distribution and        *
 * modification of SableCC.                                  *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

package org.sablecc.sablecc;

import java.util.List;
import java.util.ArrayList;
import java.util.Hashtable;

public final class DFA
{
    public DFA(NFA nfa)
    {
        this.nfa = nfa;
        construct();
        optimize();
    }

    public NFA nfa;
    public final List states = new ArrayList(10);
    public final Hashtable finder = new Hashtable(1);

    private void optimize()
    {
        List transitions = new ArrayList(states.size());
        for(int i = 0; i < states.size(); i++)
        {
            DFA.State state = (DFA.State) states.get(i);
            transitions.add(new ArrayList(0));

            for(int j = 0; j < state.transitions.size(); j++)
            {
                int max = 0;
                int st = -1;

                for(int k = 0; k < i; k++)
                {
                    int match = match(i,j,k);

                    if(match > max)
                    {
                        max = match;
                        st = k;
                    }
                }

                if(max < 2)
                {
                    ((List) transitions.get(i)).add(
                        state.transitions.get(j));
                }
                else
                {
                    DFA.Transition transition1 =
                        (DFA.Transition) state.transitions.get(j);
                    DFA.Transition transition2 =
                        (DFA.Transition) state.transitions.get(j + max - 1);

                    DFA.Transition transition =
                        new DFA.Transition(
                        new CharSet.Interval(
                        transition1.interval().start,
                        transition2.interval().end),
                        -2 - st);

                    ((List) transitions.get(i)).add(transition);
                    j += max - 1;
                }
            }
        }

        for(int i = 0; i < states.size(); i++)
        {
            DFA.State state = (DFA.State) states.get(i);
            state.transitions = (List) transitions.get(i);
        }
    }

    private int match(int st1, int tr, int st2)
    {
        DFA.State state1 = (DFA.State) states.get(st1);
        DFA.State state2 = (DFA.State) states.get(st2);

        DFA.Transition first =
            (DFA.Transition) state1.transitions.get(tr);

        int j = -1;

        for(int i = 0; i < state2.transitions.size(); i++)
        {
            DFA.Transition transition =
                (DFA.Transition) state2.transitions.get(i);

            if(transition.match(first))
            {
                j = i;
                break;
            }
        }

        if(j == -1)
        {
            return 0;
        }

        int max = 0;
        int i = tr;

        while((i < state1.transitions.size()) &&
            (j < state2.transitions.size()))
        {
            DFA.Transition transition1 =
                (DFA.Transition) state1.transitions.get(i);

            DFA.Transition transition2 =
                (DFA.Transition) state2.transitions.get(j);

            if(!transition1.match(transition2))
            {
                return max;
            }

            max++;
            i++;
            j++;
        }

        return max;
    }

    private void construct()
    {
        computeEClosures();

        IntSet initial = new IntSet();
        initial.or(eclosure(0));

        State state = new State(initial);
        states.add(state);
        finder.put(state.nfaStates, new Integer(0));

        int i = -1;
        while(++i < states.size())
        {
            System.out.print(".");

            state = (State) states.get(i);

            CharSet.Interval interval = new CharSet.Interval((char) 0, (char) 0xffff);

            do
            {
                IntSet destination = new IntSet();
                interval.end = (char) 0xffff;
                boolean modified = false;

                int[] elements = state.nfaStates.elements();

                for(int k = 0; k < elements.length; k++)
                {
                    int j = elements[k];

                    if((nfa.states[j].transition0 != null) &&
                        (nfa.states[j].transition0.chars != null))
                    {
                        CharSet.Interval overlap =
                            nfa.states[j].transition0.chars.findOverlap(interval);

                        if(overlap != null)
                        {
                            if(overlap.start > interval.start)
                            {
                                interval.end = (char) (overlap.start - 1);
                            }
                            else
                            {
                                destination.set(nfa.states[j].transition0.destination);
                                modified = true;

                                if(overlap.end < interval.end)
                                {
                                    interval.end = overlap.end;
                                }
                            }
                        }
                    }

                    if((nfa.states[j].transition1 != null) &&
                        (nfa.states[j].transition1.chars != null))
                    {
                        CharSet.Interval overlap =
                            nfa.states[j].transition1.chars.findOverlap(interval);

                        if(overlap != null)
                        {
                            if(overlap.start > interval.start)
                            {
                                interval.end = (char) (overlap.start - 1);
                            }
                            else
                            {
                                destination.set(nfa.states[j].transition1.destination);

                                if(overlap.end < interval.end)
                                {
                                    interval.end = overlap.end;
                                }
                            }
                        }
                    }
                }

                if(modified)
                {
                    destination = eclosure(destination);
                    Integer dest = (Integer) finder.get(destination);

                    if(dest != null)
                    {
                        state.transitions.add(
                            new Transition((CharSet.Interval) interval.clone(), dest.intValue()));
                    }
                    else
                    {
                        State s = new State(destination);
                        states.add(s);
                        finder.put(s.nfaStates, new Integer(states.size() - 1));

                        state.transitions.add(
                            new Transition((CharSet.Interval) interval.clone(), states.size() - 1));
                    }
                }

                interval.start = (char) (interval.end + 1);
            }
            while(interval.end != (char) 0xffff);

//            System.out.println(state);
        }

//        System.out.println(this);
    }

    private IntSet[] eclosures;

    private void computeEClosures()
    {
        eclosures = new IntSet[nfa.states.length];

        for(int i = 0; i < nfa.states.length; i++)
        {
            System.out.print(".");

            IntSet set = new IntSet();
            eclosure(i, set);
            eclosures[i] = set;
        }

        System.out.println();
    }

    private IntSet eclosure(int state)
    {
        return eclosures[state];
    }

    private void eclosure(int state, IntSet nfaStates)
    {
        if(eclosures[state] != null)
        {
            nfaStates.or(eclosures[state]);
            return;
        }

        nfaStates.set(state);

        if((nfa.states[state].transition0 != null) &&
            (nfa.states[state].transition0.chars == null) &&
            (!nfaStates.get(nfa.states[state].transition0.destination)))

        {
            eclosure(nfa.states[state].transition0.destination, nfaStates);
        }

        if((nfa.states[state].transition1 != null) &&
            (nfa.states[state].transition1.chars == null) &&
            (!nfaStates.get(nfa.states[state].transition1.destination)))

        {
            eclosure(nfa.states[state].transition1.destination, nfaStates);
        }
    }

    private IntSet eclosure(IntSet nfaStates)
    {
        IntSet result = new IntSet();

        int[] elements = nfaStates.elements();

        for(int j = 0; j < elements.length; j++)
        {
            int i = elements[j];

            result.or(eclosure(i));
        }

        return result;
    }

    public String toString()
    {
        StringBuffer result = new StringBuffer();

        for(int i = 0; i < states.size(); i++)
        {
          result.append(i).append(": ")
            .append(states.get(i))
            .append(System.getProperty("line.separator"));
        }

        return result.toString();
    }

    public final static class State
    {
        public State(IntSet nfaStates)
        {
            this.nfaStates = nfaStates;
        }

        public IntSet nfaStates = new IntSet();
        public List transitions = new ArrayList(0);
        public int accept;

        public String toString()
        {
            StringBuffer result = new StringBuffer();

            for(int i = 0; i < transitions.size(); i++)
            {
                result.append(transitions.get(i)).append(",");
            }

            return result.toString(); /*+ " " + nfaStates*/
        }
    }

    public final static class Transition
    {
        private char start;
        private char end;

        public int destination;

        public Transition(CharSet.Interval interval, int destination)
        {
            this.start = interval.start;
            this.end = interval.end;
            this.destination = destination;
        }

        public CharSet.Interval interval()
        {
            return new CharSet.Interval(start, end);
        }

        public Transition(Transition transition)
        {
            start = transition.start;
            end = transition.end;
            destination = transition.destination;
        }

        public String toString()
        {
            return destination + ":[" + interval() + "]";
        }

        public boolean match(Transition transition)
        {
            return (start == transition.start) &&
                (end == transition.end) &&
                (destination == transition.destination);
        }
    }
}


