/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * This file is part of SableCC.                             *
 * See the file "LICENSE" for copyright information and the  *
 * terms and conditions for copying, distribution and        *
 * modification of SableCC.                                  *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

package org.sablecc.sablecc;

/**/
import java.util.HashSet;
import java.util.Iterator;

/**
 * Template for executable applications. Supports command line options
 * processing model based string matching. User defined 
 * application should inherit from this base class, and implement at
 * least methods: optionsSetup() for preparing all application
 * specific command line options, constantsSetup() for application 
 * specific constant initialization and start() for the application
 * bootstrapping routines. User defined application should
 * define main() as:<br><code>
 * public static void main(String[] args){<br> 
 * ApplicationClass app = new ApplicationClass();<br>
 * app.Main(args);<br>
 * }<br></code> <br> and all necessary processing should be placed
 * inside options and start() method.
 *
 *<br><br>
 * Created: Wed Sep 22 22:22:55 1999
 *<br>
 *@author Mariusz Nowostawski
 */
public abstract class Executable {

  /** Application name. */
  protected String APP_NAME = "App";

  /** 
   * Application banner printed always after invocation. To be used
   * for short application description, copyrights, authors crap, etc.*/
  protected String APP_BANNER = "";

  /** Application banner printed always after infocation. */
  protected String APP_VERSION = "0.1";

  /** Application invocation needs to have options. */
  protected boolean NEED_OPTIONS = false;

  /**/
  private HashSet cmdln_options;
  private Option[] options;
  

  /** 
   * Main method for the template. Application main() should call
   * this method. */
  public void Main(String[] args){
    constantsSetup();
    this.options = optionsSetup();

    System.out.println(APP_NAME+"   version "+APP_VERSION);
    System.out.println("\n"+APP_BANNER+"\n");

    if(NEED_OPTIONS && args.length < 1){
      System.out.println("Please specify appropriate command line options.\n");
      displayUsage();
      System.exit(0);
    }

    bootstrap(args);
    start();
  }

  /**     
   * Application specific options declaration. User class must
   * override this  method, providing definition for all application
   * specific command-line options. This implementation returns empty
   * option table. */
  protected Option[] optionsSetup(){ 
    return new Option[0];
  }

  /**
   * Routine to postprocess the command line arguments not 
   * matched by the Options, ie. files, patterns, etc. usually 
   * added as last command line arguments.*/
  protected void processLeftOvers(Iterator enum){
    return;
  }
    
  /** 
   * User specific constant setup goes here. User class should override this
   * method. */
  protected void constantsSetup(){}

  /** 
   * Application specific code goes here. User class must override this
   * method to perform bootstrapping processing. */
  protected void start(){}

  /**
   * Usage. Should be overwritten by the application.
   */
  protected void displayUsage(){
    System.out.println(optionsSetup());
  }

  /**
   * Prepares short info for each option. This method simply
   * concatenate short help for each option with appending new line
   * character to each of them. The user has to format help lines
   * herself, or overwrite this method. */
  protected String optionsInfo(){
    StringBuffer buf = new StringBuffer();
    buf.append("Command line options:\n");
    final Option[] o = this.options;
    for(int i=0; i<o.length; i++){
      buf.append(o[i].help).append("\n");
    }
    return buf.toString();
  }

  /** 
   * Creates cmdln_options set, prepares and process the user
   * defined options. User can modify the behaviour, but for most
   * cases it should just work. */
  private void bootstrap(String[] args){
    this.cmdln_options = createOptionLookupTable(args);
    Option[] availoptions = optionsSetup();
    processOptions(availoptions);
  }
    
  /**/
  private HashSet createOptionLookupTable(String[] args){
    final HashSet options = new HashSet();
    for(int i=0; i<args.length; i++){
      options.add(args[i]);
    }
    return options;
  }
    
  /**/
  private void processOptions(Option[] option){
    for(int i=0; i<option.length; i++){
      option[i].match(this.cmdln_options);
    }

    if(this.cmdln_options.size() == 0) return;

    processLeftOvers(this.cmdln_options.iterator());

    if(this.cmdln_options.size() > 0){
      System.out.println("\nQuiting because unknown options encountered.");
      Iterator enum = this.cmdln_options.iterator();
      while(enum.hasNext()){
        System.out.println("Unknown option: "+enum.next());
      }
      System.exit(0);
    }
  }

  /**
   * Represents a single command line option. 
   * Option matching is based on pure exact string matching.
   * Each option is declared with the given key,
   * wich is matched against each token from command 
   * line options. This class does support options which takes
   * more than one token.
   **/
  public abstract class Option {

    public String[] keys;
    public String help;
    
    /** 
     * Initialize this option for a given patterns with a given help
     * string. */
    public Option(String[] keys, String help){
      this.keys = keys;
      this.help = help;
    }
    
    /**/    
    protected abstract void process(String key, Iterator iter);

    /**/
    protected void match(HashSet set){
      Iterator enum = set.iterator();
      while(enum.hasNext()){
        String s = enum.next().toString();
        for(int i=0; i<keys.length; i++){
          if(s.equals(keys[i])){ 
            enum.remove();
            process(s, enum);
          }
        }
      }
    }
  }//class Option

} // Executable
//////////////////// end of file ////////////////////
