/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * This file is part of SableCC.                             *
 * See the file "LICENSE" for copyright information and the  *
 * terms and conditions for copying, distribution and        *
 * modification of SableCC.                                  *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

package org.sablecc.sablecc;

import org.sablecc.sablecc.GenAnalyses.*;
import org.sablecc.sablecc.analysis.*;
import org.sablecc.sablecc.node.*;
import java.util.*;
import java.io.*;

public class GenTools extends DepthFirstAdapter
{
    private MacroExpander macros;
    private ResolveIds ids;
    private File pkgDir;
    private String pkgName;
    private List elemList;
    private List altList = new TypedLinkedList(AltInfoCast.instance);
    private List tokenList = new TypedLinkedList(StringCast.instance);

    ElemInfo info;

    public GenTools(ResolveIds ids)
    {
        this.ids = ids;

        try
        {
            macros = new MacroExpander(
                new InputStreamReader(
                getClass().getResourceAsStream("tools.txt")));
        }
        catch(IOException e)
        {
            throw new RuntimeException("unable to open tools.txt.");
        }

        pkgDir = new File(ids.pkgDir, "tool");
        pkgName = ids.pkgName.equals("") ? "tool" : ids.pkgName + ".tool";

        if(!pkgDir.exists())
        {
            if(!pkgDir.mkdir())
            {
                throw new RuntimeException("Unable to create " + pkgDir.getAbsolutePath());
            }
        }
    }

    public void inTokenDef1(TokenDef1 node)
    {
        tokenList.add(ids.names.get(node));
    }

    public void inAlt1(Alt1 node)
    {
        inAlt(node);
    }
 
    public void inAlt2(Alt2 node)
    {
        inAlt(node);
    }

    public void inAlt(Alt node)
    {
        elemList = new TypedLinkedList(ElemInfoCast.instance);
    }

    public void inElem1(Elem1 tnode)
    {
        info = new ElemInfo();

        info.name = (String) ids.names.get(tnode);
        info.type = (String) ids.elemTypes.get(tnode);
        info.operator = ElemInfo.NONE;

        tnode.getNode4().apply(new DepthFirstAdapter()
        {
            public void caseUnOp1(UnOp1 node)
            {
                info.operator = ElemInfo.STAR;
            }

            public void caseUnOp2(UnOp2 node)
            {
                info.operator = ElemInfo.QMARK;
            }

            public void caseUnOp3(UnOp3 node)
            {
                info.operator = ElemInfo.PLUS;
            }
        });

        elemList.add(info);
        info = null;
    }

    public void outAlt1(Alt1 node)
    {
        outAlt(node);
    }

    public void outAlt2(Alt2 node)
    {
        outAlt(node);
    }

    public void outAlt(Alt node)
    {
        AltInfo info = new AltInfo();

        info.name = (String) ids.names.get(node);
        info.elems.addAll(elemList);
        elemList = null;

        altList.add(info);
    }

    public void outStart1(Start1 node)
    {
      createDisplayTree();
      generateFile("TreeBuilder");
      generateFile("PrintWalker");
      generateFile("PrintTree");
      generateVersion();

      System.out.println();
    }

    private void generateFile(String macro)
    {
      BufferedWriter file;
      try
      {
        file = new BufferedWriter(
                  new FileWriter(
                  new File(pkgDir, macro+".java")));
        macros.apply(file, macro, new String[] { pkgName,ids.pkgName });
        file.flush();
        file.close();
      }
      catch(IOException e) 
      {
        throw new RuntimeException("Unable to create " + new File(pkgDir, macro + ".java").getAbsolutePath());
      }
      System.out.print(".");
    }


    private void generateVersion()
    {
      BufferedWriter file;
      try
      {
        file = new BufferedWriter(
                  new FileWriter(
                  new File(pkgDir, "Version.java")));
        macros.apply(file, "Version", new String[] { pkgName,ids.pkgName });
        file.flush();
        file.close();
        file = new BufferedWriter(
                  new FileWriter(
                  new File(pkgDir, "Version.java.in")));
        macros.apply(file, "Version", new String[] { pkgName,ids.pkgName });
        file.flush();
        file.close();
      }
      catch(IOException e) 
      {
        throw new RuntimeException("Unable to create " + new File(pkgDir, "Version.java").getAbsolutePath());
      }
      System.out.print(".");
    }


    private void createDisplayTree()
    {
        System.out.print(".");
        BufferedWriter file;
        
        try
        {
            file = new BufferedWriter(
                new FileWriter(
                new File(pkgDir, "DisplayTree.java")));
        }
        catch(IOException e)
        {
            throw new RuntimeException("Unable to create " + new File(pkgDir, "DisplayTree.java").getAbsolutePath());
        }

        try
        {
            macros.apply(file, "DisplayTreeHeader", new String[] {pkgName, ids.pkgName});

            for(Iterator i = altList.iterator(); i.hasNext();)
            {
                AltInfo info = (AltInfo) i.next();

                macros.apply(file, "DisplayTreeNodeHeader",
                  new String[] {info.name});

                for(Iterator j = info.elems.iterator(); j.hasNext();)
                {
                    ElemInfo eInfo = (ElemInfo) j.next();

                    switch(eInfo.operator)
                    {
                    case ElemInfo.QMARK:
                    case ElemInfo.NONE:
                        {
                          macros.apply(file, "DisplayTreeCaseBodyNode",
                                       new String[] {eInfo.name});
                        }
                        break;
                    case ElemInfo.STAR:
                    case ElemInfo.PLUS:
                        {
                          macros.apply(file, "DisplayTreeCaseBodyList",
                                       new String[] {eInfo.name, eInfo.type});
                        }
                        break;
                    }
                }

                macros.apply(file, "DisplayTreeNodeFooter",
                             new String[] {info.name});

            }

            macros.apply(file, "DisplayTreeFooter", null);
        }
        catch(IOException e)
        {
            throw new RuntimeException("An error occured while writing to " +
                new File(pkgDir, "DisplayTree.java").getAbsolutePath());
        }
        catch(Exception e){
          e.printStackTrace();
        }

        try
        {
            file.flush();
            file.close();
        }
        catch(IOException e)
        {
        }

    } 

    private static class AltInfoCast implements Cast
    {
        final static AltInfoCast instance = new AltInfoCast();

        private AltInfoCast()
        {
        }

        public    Object cast(Object o)
        {
            return (AltInfo) o;
        }
    }

    private static class ElemInfo
    {
        final static int NONE = 0;
        final static int STAR = 1;
        final static int QMARK = 2;
        final static int PLUS = 3;

        String name;
        String type;
        int operator;
    }

    private static class ElemInfoCast implements Cast
    {
        public final static ElemInfoCast instance = new ElemInfoCast();

        private ElemInfoCast()
        {
        }

        public    Object cast(Object o)
        {
            return (ElemInfo) o;
        }
    }

    private static class AltInfo
    {
        String name;
        final List elems = new TypedLinkedList(ElemInfoCast.instance);
    }
}


