/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * This file is part of SableCC.                             *
 * See the file "LICENSE" for copyright information and the  *
 * terms and conditions for copying, distribution and        *
 * modification of SableCC.                                  *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

package org.sablecc.sablecc;

import java.io.*;
import java.awt.*;
import java.util.*;
import org.sablecc.sablecc.node.*;
import org.sablecc.sablecc.analysis.*;
import org.sablecc.sablecc.lexer.*;
import org.sablecc.sablecc.parser.*;
import java.util.Vector;

public class SableCC extends Executable
{
  Vector files = new Vector();
  String destinationPath = ".";
  boolean withTools = false;
  
  
  /** Application banner, name and version.*/
  protected void constantsSetup()
  {
    APP_NAME = "SableCC";
    APP_VERSION = Version.VERSION;
    APP_BANNER = 
      "Copyright (C) 1997-2001 Etienne M. Gagnon <egagnon@j-meg.com> and others."+
      "\nAll rights reserved.\n"+
      "\n"+
      "This software comes with ABSOLUTELY NO WARRANTY.  This is free software,\n"+
      "and you are welcome to redistribute it under certain conditions.\n"+
      "\n"+
      "Type 'sablecc --license' to view\n"+
      "the complete copyright notice and license.\n";
    NEED_OPTIONS = true;
  }


  protected void displayUsage()
  {
    System.out.println("Usage:");
    System.out.println("   java -jar sablecc.jar [-h|--help] [-l|--license]");
    System.out.println("   java -jar sablecc.jar [-d path] [--with-tools] grammar1 grammar2 ...");
    System.out.println("");
    System.out.println(optionsInfo());
  }

  /** Application command line options. */
  protected Option[] optionsSetup()
  {

    /** Common processing unit for -d --destination and -o --outputfile. */
    final Option processDestination_d = 
       new Option(new String[]{"-d", "--destination"}, 
                 "   -d path | --destination path        sets the destination directory for generated files")
        {
          protected void process(String key, Iterator iter)
          {
            if(iter.hasNext())
              {
                destinationPath = iter.next().toString();
                iter.remove();
              } 
            else 
              {
                System.out.println("ERROR:  Missing destination directory!\n\n");
                displayUsage();
                System.exit(1);
              }
          }
         };

    final Option processDestination_o = 
      new Option(new String[]{"-o", "--outputdirectory"}, 
                 "   -o path | --outputdirectory path    same as -d and --destination")
        {
          protected void process(String key, Iterator iter)
          {
            processDestination_d.process(key, iter);
          }
        };

    return new Option[]
    { 
      new Option(new String[] {"-l", "--license"}, 
                 "   -l      | --license                 displays copyright notice and license")
        {
          protected void process(String key, Iterator iter)
          {
            new DisplayLicense();
          }
        },
      new Option(new String[]{"-h", "--help"}, 
                 "   -h      | --help                    displays command line options")
        {
          protected void process(String key, Iterator iter)
          {
            displayUsage();
            System.exit(1);
          }
        },
      processDestination_d,
      processDestination_o,
      new Option(new String[]{"-t", "--with-tools"},
                 "   -t      | --with-tools              will generate tools for the generated parser")
        {
          protected void process(String key, Iterator iter)
          {
            withTools = true;
          }
        }
    };
  }


  /** Lets treat all the remaining command line arguments as grammar files. */
  protected void processLeftOvers(Iterator iter)
  {
    while(iter.hasNext())
      {
        files.addElement(iter.next().toString());
        iter.remove();
      }
  }


  /** 
   * Main working horse. All is ready to be processed. 
   * files contain all the grammar files
   * destinationPath is set to . or to the path
   * withTools is either true or false
   */
  protected void start()
  {
    try
      {
        for(int i=0; i<files.size(); i++)
          {
            processGrammar((files.elementAt(i)).toString(), destinationPath, withTools);
          }
      }
    catch(Exception e)
      {
        System.out.println(e);
        System.exit(1);
      }
    catch(Throwable e)
      {
        System.out.println(e);
        System.exit(1);
      }
    finally
      {
        System.exit(0);
      }
  }


  /**
   * The main method for processing grammar file and generating the parser/lexer.
   * @param in input grammar file
   * @param dir output directory
   */
  public static void processGrammar(String grammar, String destDir, boolean withTools) 
    throws Exception, Throwable
  {
    File in;
    File dir;
    
    in = new File(grammar);
    in = new File(in.getAbsolutePath());
    
    if(destDir == null)
      {
        dir = new File(in.getParent());
      }
    else
      {
        dir = new File(destDir);
        dir = new File(dir.getAbsolutePath());
      }

    processGrammar(in, dir, withTools);
  }

  /**
   * The main method for processing grammar file and generating the parser/lexer.
   * @param in input grammar file
   * @param dir output directory
   */
  public static void processGrammar(File in,  File dir, boolean withTools) 
    throws Exception, Throwable
  {
    if(!in.exists())
      { 
        System.out.println("ERROR: grammar file "+in.getName()+" does not exist."); 
        System.exit(1); 
      }

    if(!dir.exists())
      { 
        System.out.println("ERROR: destination directory "+dir.getName()+" does not exist."); 
        System.exit(1); 
      }

    // re-initialize all static structures in the engine
    LR0Collection.reinit();
    Symbol.reinit();
    Production.reinit();
    Grammar.reinit();

    System.out.println("\n -- Generating parser for "+in.getName()+" in "+dir.getPath());
    
    FileReader temp;
    
    // Build the AST
    Start tree = new Parser(new Lexer(new BufferedReader(temp = new FileReader(in)))).parse();
    temp.close();
    
    System.out.println("Verifying identifiers.");
    ResolveIds ids = new ResolveIds(dir);
    tree.apply(ids);
    
    // Create the node.* and analysis.* files
    System.out.println("Generating token classes.");
    tree.apply(new GenTokens(ids));
    
    System.out.println("Generating production classes.");
    tree.apply(new GenProds(ids));
    
    System.out.println("Generating alternative classes.");
    tree.apply(new GenAlts(ids));

    System.out.println("Generating analysis classes.");
    tree.apply(new GenAnalyses(ids));

    System.out.println("Generating utility classes.");
    tree.apply(new GenUtils(ids));
      
    try
      {
        System.out.println("Generating the lexer.");
        tree.apply(new GenLexer(ids));
      }
    catch(Exception e)
      {
        System.out.println(e.getMessage());
        throw e;
      }

    try
      {
        System.out.println("Generating the parser.");
        tree.apply(new GenParser(ids));
      }
    catch(Exception e)
      {
        System.out.println(e.getMessage());
        throw e;
      }

    if(withTools) 
      {
        try
          {
            System.out.println("Generating user tools for the generated parser.");
            tree.apply(new GenTools(ids));
          }
        catch(Exception e)
          {
            System.out.println(e.getMessage());
            throw e;
          }
      }
  }


  public static void main(String[] arguments)
  {
    SableCC app = new SableCC();
    app.Main(arguments);
  }

}
